#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <pthread.h>
#include "sock.h"
#include "client.h"
#include "proto.h"
#include "geom.h"
#include "galton.h"
#include "timer.h"

#if defined(WIN32) || defined(__WIN32__)
#define close(s)	closesocket(s)

#define F_GETFL		0
#define F_SETFL		1

#define O_NONBLOCK	1
#define EINPROGRESS	WSAEWOULDBLOCK
#define ETIMEDOUT	WSAETIMEDOUT

typedef int socklen_t;

static int fcntl(int fd, int cmd, ...);
#endif


static void *client_thread(void *data);
static bool recv_greeting(int s);
static void bounds_callback(const SimBall *ball, void *cls);

static pthread_t tid;
static pthread_mutex_t sock_mutex = PTHREAD_MUTEX_INITIALIZER;
static int sock = -1;
static volatile bool client_done;

// saved for the retry loop
static const char *saved_hostname;
static int saved_port;

bool client_connect(const char *hostname, int port)
{
	int s, flags = 0;
	int conn_timeout = 0;

	cfg.getopt_int("conn_timeout", &conn_timeout);
	bool nonblock = conn_timeout > 0;

	saved_hostname = hostname;
	saved_port = port;

	if((s = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		perror("failed to create socket");
		return false;
	}


	if(nonblock) {
		// put the socket in non-blocking mode
		flags = fcntl(s, F_GETFL, 0);
		if(fcntl(s, F_SETFL, flags | O_NONBLOCK) == -1) {
			fprintf(stderr, "failed to set non-blocking, connect might take long to timeout\n");
		}
	}

	struct hostent *host = gethostbyname(hostname);
	if(!host) {
		fprintf(stderr, "%s: %s\n", hstrerror(h_errno), hostname);
		close(s);
		return false;
	}

	struct sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr = *((struct in_addr*)host->h_addr);

	int res;
	if((res = connect(s, (struct sockaddr*)&addr, sizeof addr) != 0)) {
#if defined(WIN32) || defined(__WIN32__)
		errno = WSAGetLastError();
#endif
		if(errno != EINPROGRESS) {
			fprintf(stderr, "failed to connect to: %s: %s\n", hostname, strerror(errno));
			close(s);
			return false;
		}
	}

	/* if we're doing a non-blocking connect, we need to call select to determine
	 * when the connection initiation is ready.
	 */
	if(nonblock && res != 0) {
		fd_set rset, wset, eset;
		FD_ZERO(&rset);
		FD_ZERO(&wset);
		FD_ZERO(&eset);
		FD_SET(s, &rset);
		FD_SET(s, &wset);
		FD_SET(s, &eset);

		do {
			struct timeval tv;
			tv.tv_sec = conn_timeout;
			tv.tv_usec = 0;
			res = select(s + 1, &rset, &rset, &eset, &tv);
		} while(res == -1 && errno == EINTR);

		if(!res) {
			fprintf(stderr, "failed to connect to: %s: %s\n", hostname, strerror(ETIMEDOUT));
			close(s);
			return false;
		}

		if(FD_ISSET(s, &rset) || FD_ISSET(s, &wset)) {
			int err;
			socklen_t len = sizeof err;

			if(getsockopt(s, SOL_SOCKET, SO_ERROR, (char*)&err, &len) == -1 || err) {
				fprintf(stderr, "error while connecting: %s\n", strerror(err));
				close(s);
				return false;
			}
		} else {
			fprintf(stderr, "WTF?\n");
			return false;
		}
	}

	fcntl(s, F_SETFL, flags);


	// verify we're connected to the correct server
	if(!recv_greeting(s)) {
		close(s);
		return false;
	}

	sock = s;
	if(pthread_create(&tid, 0, client_thread, 0) == -1) {
		perror("failed to create thread\n");
		close(s);
		return false;
	}

	return true;
}

void client_disconnect()
{
	if(sock == -1) return;

	MsgHeader hdr = {MSG_BYE, 0};
	
	pthread_mutex_lock(&sock_mutex);
	send_msg_header(sock, &hdr);
	pthread_mutex_unlock(&sock_mutex);
	client_done = true;

	printf("waiting for client thread to exit ...");
	fflush(stdout);

	pthread_join(tid, 0);
	printf("done\n");

	close(sock);
	sock = -1;
}

bool send_collect_msg(int slot)
{
	if(sock == -1) {
		return false;
	}

	printf("sending collect msg, slot: %d\n", slot);

	MsgHeader hdr = {MSG_COLLECT, 1};
	unsigned char data = (unsigned char)slot;

	int res = 0;
	pthread_mutex_lock(&sock_mutex);
	if(send_msg_header(sock, &hdr)) {
		res = send(sock, (const char*)&data, 1, 0);
	}
	pthread_mutex_unlock(&sock_mutex);

	return res == 1;
}

static void *client_thread(void *data)
{
	unsigned long ping_sent = 0;

	/* modify the simulation limits, to get rid of the balls when they go
	 * to the bottom half, and get a callback called so that we may pass them
	 * to the client.
	 */
	double lim_upper, lim_lower;
	double half_point = box_param.rows * box_param.row_dist / 2.0;
	sim->get_limits(&lim_lower, &lim_upper);
	sim->set_limits(lim_lower, half_point);
	sim->set_bounds_callback(bounds_callback, 0);

	while(!client_done) {
		MsgHeader hdr;
		unsigned char data_buf[256];

		if(ping_sent) {
			if(get_msec() - ping_sent > PING_TIMEOUT) {
				printf("ping timeout...\n");
				break;
			}
		} else {
			pthread_mutex_lock(&sock_mutex);
			printf("ping...\n");
			ping(sock);
			pthread_mutex_unlock(&sock_mutex);

			ping_sent = get_msec();
		}

		// get message
		if(!recv_msg_header(sock, &hdr, TIMEOUT)) {
			continue;
		}
		// get the message data into the data buffer
		if(recv_data(sock, data_buf, hdr.data_sz, TIMEOUT) < hdr.data_sz) {
			printf("huh?\n");
			continue;
		}

		printf("got message: %s\n", get_msg_name(hdr.msg));

		switch(hdr.msg) {
		case MSG_ALIVE:
			ping_sent = 0;
			break;

		case MSG_SYNC:
			pthread_mutex_lock(&ball_stats.mutex);
			parse_sync_data(data_buf, &ball_stats);
			pthread_mutex_unlock(&ball_stats.mutex);
			break;

		case MSG_SPAWN:
			pthread_mutex_lock(&ball_stats.mutex);
			ball_stats.num_spawned++;
			pthread_mutex_unlock(&ball_stats.mutex);
			break;

		case MSG_PASS_BALL:
			{
				SimBall *ball = new SimBall;
				parse_ball_data(data_buf, ball);

				ball->rad = get_ball_radius(ball->type);
				ball->mass = get_ball_mass(ball->type);
				ball->rest = get_ball_restitution(ball->type);
				sim->add_ball(ball);
			}
			break;

		case MSG_PING:
			pthread_mutex_lock(&sock_mutex);
			alive(sock);
			pthread_mutex_unlock(&sock_mutex);
			break;

		default:
			fprintf(stderr, "unexpected message: %s\n", get_msg_name(hdr.msg));
		}
	}

	// if the connection was interrupted, retry indefinitely
	int num_try = 0;
	while(!client_done) {
		if(client_connect(saved_hostname, saved_port)) {
			pthread_detach(pthread_self());
			break;
		}

		num_try++;
		int sleep_msec = MIN(num_try, 10) * 1000;
		fprintf(stderr, "will try again in: %d sec\n", sleep_msec / 1000);
		msleep(sleep_msec);
	}

	return 0;
}

static bool recv_greeting(int s)
{
	MsgHeader hdr;

	if(!recv_msg_header(s, &hdr, TIMEOUT)) {
		fprintf(stderr, "timed out while waiting for a greeting from the server\n");
		return false;
	}

	if(hdr.msg != MSG_HELLO) {
		fprintf(stderr, "protocol error, server didn't greet me properly...\n");
		return false;
	}
	char *buf = (char*)alloca(hdr.data_sz + 1);

	if(recv_data(s, buf, hdr.data_sz, TIMEOUT) != hdr.data_sz) {
		fprintf(stderr, "timed out while waiting for greeting text\n");
		return false;
	}
	buf[hdr.data_sz] = 0;

	if(strstr(buf, "GALTON") != buf) {
		fprintf(stderr, "protocol error, server didn't greet me properly...\n");
		return false;
	}

	printf("connected, server greeting: %s\n", buf);
	return true;
}

/* This function is called by Simulation whenever a ball escapes the specified
 * upper or lower limits. Our task is to see if that was the upper limit (half-way
 * point of the box) and if so, pass the ball to the server.
 */
static void bounds_callback(const SimBall *ball, void *cls)
{
	double upper, lower;
	sim->get_limits(&lower, &upper);

	if(ball->pos.y > upper) {
		printf("passing ball: [%.3f, %.3f, %.3f]\n", ball->pos.x, ball->pos.y, ball->pos.z);

		pthread_mutex_lock(&sock_mutex);
		pass_ball(sock, ball);
		pthread_mutex_unlock(&sock_mutex);
	}
}

#if defined(WIN32) || defined(__WIN32__)
static int fcntl(int fd, int cmd, ...)
{
	va_list ap;
	int res;

	va_start(ap, cmd);
	switch(cmd) {
	case F_GETFL:
		{
			int *ptr = va_arg(ap, int*);
			if(ptr) *ptr = 0;
			res = 0;
		}
		break;

	case F_SETFL:
		{
			int val = va_arg(ap, int);
			unsigned long uval = val;
			res = ioctlsocket(fd, FIONBIO, &uval);
		}
		break;

	default:
		res = -1;
	}
	va_end(ap);

	return res;
}
#endif
