#if defined(WIN32) || defined(__WIN32__)
#include <malloc.h>
#else
#include <alloca.h>
#endif

#include <henge.h>
#include "draw.h"
#include "geom.h"
#include "text.h"
#include "galton.h"

using namespace henge;

// draws all balls by accessing the ball visualization data in Simulation.
void draw_balls(unsigned int time, void *cls)
{
	Simulation *sim = (Simulation*)cls;

	/* make a private copy to avoid holding the mutex while we render
	 * an delaying the simulation loop
	 */
	pthread_mutex_lock(&sim->ballvis_mutex);
	int num_balls = sim->num_vis_balls;

	if(!num_balls) {
		pthread_mutex_unlock(&sim->ballvis_mutex);
		return;
	}

	BallVis *bvis = (BallVis*)alloca(num_balls * sizeof *bvis);
	memcpy(bvis, &sim->ball_vis[0], num_balls * sizeof *bvis);
	pthread_mutex_unlock(&sim->ballvis_mutex);
	
	for(int i=0; i<num_balls; i++) {
		if(bvis[i].type == BTYPE_INVALID) {
			continue;
		}
		robject *obj = get_ball_object(bvis[i].type);
		obj->set_position(bvis[i].pos);
		obj->render();
	}
}

/* custom render function (henge::robject::set_render_func) for the metal balls.
 * it uses the fixed function texgen to render a reflection of the static
 * sky cubemap on the balls.
 */
void envball_render(const robject *obj, unsigned int msec, void *cls)
{
	glMatrixMode(GL_MODELVIEW);
	Matrix4x4 view_matrix;
	store_matrix(&view_matrix);
	view_matrix = view_matrix.inverse();
	view_matrix[0][3] = view_matrix[1][3] = view_matrix[2][3] = 0.0;
	view_matrix[3][0] = view_matrix[3][1] = view_matrix[3][2] = 0.0;
	view_matrix[3][3] = 1.0;
	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	load_matrix(view_matrix);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	mult_matrix(obj->get_xform_matrix(msec));

	glPushAttrib(GL_COLOR_BUFFER_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT);

	const material *mat = &obj->get_material();
	mat->bind();

	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
	glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glEnable(GL_TEXTURE_GEN_R);
	glassert();

	set_texenv_color(0, GL_ADD, GL_PREVIOUS, GL_TEXTURE);
	glassert();

	obj->get_mesh()->draw();

	set_default_texenv(0);

	if(mat && mat->get_shader()) {
		set_shader(0);
	}

	glPopAttrib();

	glMatrixMode(GL_TEXTURE);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

/* custom render function (henge::robject::set_render_func) for the skydome.
 * it uses the fixed function texgen to map the sky cubemap onto a surrounding
 * low-poly sphere.
 */
void sky_render(const robject *obj, unsigned int msec, void *cls)
{
	glMatrixMode(GL_MODELVIEW);
	Matrix4x4 view_matrix;
	store_matrix(&view_matrix);
	view_matrix = view_matrix.inverse();
	view_matrix[0][3] = view_matrix[1][3] = view_matrix[2][3] = 0.0;
	view_matrix[3][0] = view_matrix[3][1] = view_matrix[3][2] = 0.0;
	view_matrix[3][3] = 1.0;
	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	load_matrix(view_matrix);

	glPushAttrib(GL_COLOR_BUFFER_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT);

	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB);
	glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB);
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glEnable(GL_TEXTURE_GEN_R);

	glDisable(GL_LIGHTING);
	glColor3f(1, 1, 1);

	const material *mat = &obj->get_material();
	mat->bind();

	glFrontFace(GL_CW);
	obj->get_mesh()->draw();
	glFrontFace(GL_CCW);

	if(mat && mat->get_shader()) {
		set_shader(0);
	}
	
	glPopAttrib();

	glMatrixMode(GL_TEXTURE);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void draw_slot_counter(unsigned int msec, void *cls)
{
	static int cols = -1;
	static float pin_dist, pin_length;

	if(cols < 0) {
		cols = DEF_COLS;
		pin_dist = DEF_PIN_DIST;
		pin_length = DEF_PIN_LEN;

		cfg.getopt_int("columns", &cols);
		cfg.getopt("pin_dist", &pin_dist);
		cfg.getopt("pin_length", &pin_length);
	}

	int num = *(int*)cls;
	char buf[16];
	sprintf(buf, "%d", num);

	int slot_idx = (int*)cls - ball_stats.exit_counters;
	double x = (double)slot_idx * pin_dist - pin_dist * (double)(cols - 2) / 2.0 + pin_dist / 2.0;

	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);
	glDepthMask(0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(x, 0, pin_length / 2.0 + 0.01);
	glScalef(60, 60, 60);

	set_text_pos(0.5 - get_text_width(buf) / 2.0, 0.5);
	set_text_color(1, 1, 1, 1);
	set_text_mode(TEXT_MODE_3D);

	print_string(buf);

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glPopAttrib();
	glDepthMask(1);
}

#define TEXTCOL()		set_text_color(0.7, 0.7, 0.7, 1)
#define BLUECOL()		set_text_color(0.34, 0.44, 0.5, 1)
#define GREENCOL()		set_text_color(0.4, 0.65, 0.4, 1)
#define REDCOL()		set_text_color(0.6, 0, 0, 1)

void draw_stats()
{
	char buf[16];
	int bin = ball_stats.num_spawned;
	int bout = ball_stats.num_collected;
	int missing = bin - bout;

	set_text_pos(0.01, 0.05);
	set_text_mode(TEXT_MODE_2D);
	set_text_size(0.75);

	TEXTCOL();
	print_string("Balls collected: ");
	
	sprintf(buf, "%d ", bout);
	BLUECOL();
	print_string(buf);
	
	TEXTCOL();
	print_string("/ ");

	BLUECOL();
	sprintf(buf, "%d ", bin);
	print_string(buf);

	set_text_pos(0.4, 0.05);
	TEXTCOL();
	print_string("Missing: ");

	sprintf(buf, "%d", missing);
	if(missing != 0) {
		REDCOL();
	} else {
		GREENCOL();
	}
	print_string(buf);

	set_text_color(1, 1, 1, 1);
	set_text_size(1.0);
}
