#include <list>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <henge.h>

#ifdef USE_GXBASE
#include "gxglut.h"
#else

#if defined(__APPLE__) && defined(__MACH__)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#endif	// USE_GXBASE

#include "galton.h"
#include "geom.h"
#include "sim.h"
#include "draw.h"
#include "text.h"
#include "procgraph.h"
#include "timer.h"
#include "server.h"
#include "client.h"

#define GRAVITY		-980

using namespace std;
using namespace henge;

static bool init(int argc, char **argv);
static void cleanup();
static void disp();
static void debug_draw();
static void redisplay();
static void reshape(int x, int y);
static void keyb(unsigned char key, int x, int y);
static void mouse(int bn, int state, int x, int y);
static void motion(int x, int y);

static float cam_theta = 0, cam_phi = 5, cam_dolly = 200;
static float cam_x, cam_y = 60;

static scene *scn;
Simulation *sim;

static float timestep = 0.0025;
static unsigned long frame_interval;	// minimum interval between frame updates

config_file cfg;

int xsz = 800;
int ysz = 600;
static bool fullscr = false;

static unsigned int font;

static int drop_batch = 1;
static bool show_sim_obj = false;
static bool show_border = true;
static bool show_usage = false;
static texture *border_tex;

static unsigned int drop_type = BTYPE_RUBBER;

static ProcGraph *pgraph;
static int num_threads;

static bool server_mode = true;


int main(int argc, char **argv)
{
#if !defined(WIN32) && !defined(__WIN32__)
	signal(SIGPIPE, SIG_IGN);
#endif

	if(!init(argc, argv)) {
		return 1;
	}
	atexit(cleanup);

	glutMainLoop();
	return 0;
}

static bool init(int argc, char **argv)
{
	const char *srv_host = "localhost";
	int srv_port = 1234;
	
	cfg.read("game.cfg");
	
	cfg.getopt_int("xres", &xsz);
	cfg.getopt_int("yres", &ysz);

	cfg.getopt("fullscreen", &fullscr);
	cfg.getopt("timestep", &timestep);

	int fps = 30;
	cfg.getopt_int("fps", &fps);
	frame_interval = 1000 / fps;

	glutInitWindowSize(xsz, ysz);
	glutInit(&argc, argv);

	for(int i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			switch(argv[i][1]) {
			case 's':
				server_mode = true;
				break;

			case 'c':
				if(argv[i + 1] && argv[i + 1][0] != '-') {
					srv_host = argv[++i];
				}
				server_mode = false;
				break;

			case 'p':
				if(!isdigit(argv[++i][0])) {
					fprintf(stderr, "-p must be followed by a port number\n");
					return false;
				}
				srv_port = atoi(argv[i]);
				break;

			default:
				fprintf(stderr, "unrecognized argument: %s\n", argv[i]);
			case 'h':
				printf("Usage: %s [-s|-c <host|ip>] [-p <port>] [-h]\n", argv[0]);
				return argv[i][1] == 'h' ? 0 : 1;
			}
		} else {
			fprintf(stderr, "unexpected argument: %s\n", argv[i]);
			fprintf(stderr, "Usage: %s [-s|-c <host|ip>] [-p <port>] [-h]\n", argv[0]);
			return false;
		}
	}

	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);

	if(fullscr) {
		char game_mode_str[32];
		sprintf(game_mode_str, "%dx%d", xsz, ysz);
		glutGameModeString(game_mode_str);
		glutEnterGameMode();
	} else {
		glutCreateWindow("Galton Box");
	}

	glutDisplayFunc(disp);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutIdleFunc(redisplay);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);

	get_msec();	// reset timer

	if(!henge::init()) {
		return false;
	}
	scn = new scene;

	scn->add_light(new light(Vector3(200, 200, 400)));

	sim = new Simulation;

	if(!init_geom(scn, sim)) {
		return false;
	}

	float drag_coef = 0.0;
	cfg.getopt("drag", &drag_coef);

	// distances are in centimeters so gravity is -0.098cm/s^2
	sim->set_gravity(Vector3(0.0, GRAVITY, 0.0));
	sim->set_drag(drag_coef);
	sim->set_timeslice(timestep);

	// load font
	if(!(font = create_font("data/kerkis.pfb", 32))) {
		fprintf(stderr, "failed to load font\n");
		return false;
	}

	cfg.getopt("border", &show_border);
	border_tex = get_texture("border.png");

	// initialize utilization graphs
	num_threads = sim->tpool.get_num_threads();
	pgraph = new ProcGraph[num_threads];

	Vector2 pos(0.9 * (float)xsz / (float)ysz, 0.87);
	Vector2 sz(0.2, 0.1);

	for(int i=0; i<num_threads; i++) {
		pgraph[i].init(i, &sim->tpool, 64);
		pgraph[i].set_position(pos);
		pgraph[i].set_size(sz);

		pos.y -= sz.y * 1.1;
	}

	int port = 1234;
	cfg.getopt_int("port", &port);
	
	if(server_mode) {
		if(!init_server(port)) {
			fprintf(stderr, "server init failed, won't be able to accept connections\n");
		}
	} else {
		if(!client_connect(srv_host, port)) {
			return false;
		}
	}

	// start the physics simulation thread
	sim->start();
	return true;
}

static void cleanup()
{
	sim->stop();

	shutdown_server();
	client_disconnect();
	
	delete [] pgraph;
	delete_font(font);
	delete scn;
	delete sim;

	henge::shutdown();
}

static void disp()
{
	unsigned long msec = get_msec();

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(cam_x, -cam_y, -cam_dolly);
	glRotatef(cam_phi, 1, 0, 0);
	glRotatef(cam_theta, 0, 1, 0);

	scn->render();

	debug_draw();

	if(show_border) {
		overlay(border_tex, color(0.34, 0.44, 0.5, 0.75));
	}
	draw_stats();

	static unsigned long prev_stats_upd;
	bool do_upd = false;
	if(msec - prev_stats_upd >= 500) {
		do_upd = true;
		prev_stats_upd = msec;
	}

	for(int i=0; i<num_threads; i++) {
		if(do_upd) {
			pgraph[i].update();
		}
		if(show_usage) {
			pgraph[i].draw();
		}
	}

	glutSwapBuffers();

	unsigned long frame_dur = get_msec() - msec;
	if(frame_dur < frame_interval) {
		msleep(frame_interval - frame_dur);
	}
}

void debug_draw()
{
	if(show_sim_obj) {
		sim->draw_static_objects(color(1.0, 0.4, 0.3, 0.5));
	}

	if(!sim->col_acc_list.empty()) {
		glPushAttrib(GL_ENABLE_BIT);
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);
		glPointSize(2.0);

		glBegin(GL_POINTS);
		glColor3f(1, 0, 0);

		list<Contact>::iterator iter = sim->col_acc_list.begin();
		while(iter != sim->col_acc_list.end()) {
			Vector3 v = (iter++)->pos;
			glVertex3f(v.x, v.y, v.z);
		}
		glEnd();

		glPopAttrib();
		sim->col_acc_list.clear();
	}

	if(!sim->cont_acc_list.empty()) {
		glPushAttrib(GL_ENABLE_BIT);
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);
		glPointSize(2.0);

		glBegin(GL_POINTS);
		glColor3f(1, 1, 0);

		list<Contact>::iterator iter = sim->cont_acc_list.begin();
		while(iter != sim->cont_acc_list.end()) {
			Vector3 v = (iter++)->pos;
			glVertex3f(v.x, v.y, v.z);
		}
		glEnd();

		glPopAttrib();
		sim->cont_acc_list.clear();
	}
}

static void redisplay()
{
	glutPostRedisplay();
}

static void reshape(int x, int y)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (float)x / (float)y, 1.0, 10000.0);

	glViewport(0, 0, x, y);
}

static void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
	case 'x':
		exit(0);

	case 'w':
		{
			static bool wind;
			wind = !wind;
			if(wind) {
				sim->set_gravity(Vector3(1000, GRAVITY, 0));
			} else {
				sim->set_gravity(Vector3(0, GRAVITY, 0));
			}
		}
		break;

	case 'W':
		{
			static bool wire;
			wire = !wire;
			if(wire) {
				glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			} else {
				glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
			}
			glutPostRedisplay();
		}
		break;

	case 'c':
		show_sim_obj = !show_sim_obj;
		glutPostRedisplay();
		break;

	case ' ':
		{
			static bool slow_motion;
			slow_motion = !slow_motion;
			if(slow_motion) {
				sim->set_time_scaling(0.1);
			} else {
				sim->set_time_scaling(1.0);
			}
		}
		break;
	
	case '=':
		if(++drop_batch > 100) {
			drop_batch = 100;
		}
		break;

	case '-':
		if(--drop_batch < 0) {
			drop_batch = 0;
		}
		break;

	case '\b':
	case 127:
		if(server_mode) {
			drop_balls(sim, drop_batch, drop_type);
		}
		break;

	case 's':
		sim->toggle_pause();
		if(sim->is_paused()) {
			glutIdleFunc(0);
		} else {
			glutIdleFunc(redisplay);
		}
		break;

	case '1':
	case '2':
	case '3':
	case '4':
		drop_type = key - '1';
		break;

	case '[':
		{
			static bool col_enabled;
			col_enabled = !col_enabled;
			sim->set_collision_gathering(col_enabled);
			if(!col_enabled) {
				sim->col_acc_list.clear();
			}
		}
		break;

	case ']':
		{
			static bool cont_enabled;
			cont_enabled = !cont_enabled;
			sim->set_contact_gathering(cont_enabled);
			if(!cont_enabled) {
				sim->cont_acc_list.clear();
			}
		}
		break;

	case 'u':
		show_usage = !show_usage;
		break;
	}
}

static int bnstate[16];

static int prev_x = -1, prev_y;
static void mouse(int bn, int state, int x, int y)
{
	bnstate[bn] = state == GLUT_DOWN ? 1 : 0;
	if(state == GLUT_DOWN) {
		if(bn == 3) {
			cam_dolly -= 20;
			glutPostRedisplay();
			if(cam_dolly < 0) cam_dolly = 0;
		} else if(bn == 4) {
			cam_dolly += 20;
			glutPostRedisplay();
		} else {
			prev_x = x;
			prev_y = y;
		}
	} else {
		prev_x = -1;
	}
}

static void motion(int x, int y)
{
	if(bnstate[0]) {
		cam_theta += (x - prev_x) * 0.5;
		cam_phi += (y - prev_y) * 0.5;

		if(cam_phi < -90) cam_phi = -90;
		if(cam_phi > 90) cam_phi = 90;

		glutPostRedisplay();
	}

	if(bnstate[1]) {
		cam_x += (prev_x - x) * 0.5;
		cam_y += (prev_y - y) * 0.5;
		glutPostRedisplay();
	}

	if(bnstate[2]) {
		cam_dolly += (y - prev_y) * 1.0;
		glutPostRedisplay();
	}

	prev_x = x;
	prev_y = y;
}
