/* geom.cc - construction of the renderable geometry and simulation objects
 * for the galton box and its surrounding environment.
 */
#include "geom.h"
#include "draw.h"
#include "galton.h"
#include "client.h"
#include "server.h"

using namespace henge;

// various geometry/simobj construction functions
static bool create_sky(scene *scn);
static bool create_ground(scene *scn, Simulation *sim);
static bool create_galton(scene *scn, Simulation *sim);
static bool create_galton_glass(scene *scn, Simulation *sim);
static bool create_galton_pins_sides(scene *scn, Simulation *sim);
static bool create_balls(scene *scn, Simulation *sim);
static bool create_exit_triggers(scene *scn, Simulation *sim);

// custom collision callback for the exit slots
static bool ball_exit_callback(const SimStaticObj *sobj, const SimBall *ball,
		const Contact *cont, void *cls);

// encapsulation and data hiding in its finest
static robject *ball[NUM_BALL_TYPES];
static float ball_mass[NUM_BALL_TYPES] = {0.1, 0.15, 0.3};
static float ball_rest[NUM_BALL_TYPES] = {0.95, 0.75, 0.4};
static float ball_rad = DEF_BALL_RAD;

// these two are calculated from the above
static BoxParameters *bp;

BallStats ball_stats;
BoxParameters box_param = BOX_PARAM_INITIALIZER;
SimStaticObj *ground_sim_obj;


/* -- init_geom --
 * reads config file options, and calls all the geometry/simobj construction
 * functions to generate the whole galton box & scenery.
 *
 * That's the only construction function visible from outside this file and is
 * called once at the beginning of the program (see galton.cc).
 */
bool init_geom(scene *scn, Simulation *sim)
{
	// datafile path
	add_path("data");

	bp = &box_param;

	cfg.getopt_int("rows", &bp->rows);
	cfg.getopt_int("columns", &bp->cols);
	cfg.getopt("pin_rad", &bp->pin_rad);
	cfg.getopt("pin_length", &bp->pin_length);
	cfg.getopt("row_dist", &bp->row_dist);
	cfg.getopt("pin_dist", &bp->pin_dist);
	cfg.getopt("ball_rad", &ball_rad);
	
	bp->height = (double)bp->rows * bp->row_dist + 2.0 * bp->pin_rad;
	bp->width = (double)(bp->cols - 1) * bp->pin_dist + 2.0 * bp->pin_rad;

	cfg.getopt("mass_rubber", &ball_mass[BTYPE_RUBBER]);
	cfg.getopt("mass_wood", &ball_mass[BTYPE_WOOD]);
	cfg.getopt("mass_steel", &ball_mass[BTYPE_METAL]);

	cfg.getopt("e_rubber", &ball_rest[BTYPE_RUBBER]);
	cfg.getopt("e_wood", &ball_rest[BTYPE_WOOD]);
	cfg.getopt("e_steel", &ball_rest[BTYPE_METAL]);

	if(!create_sky(scn)) {
		return false;
	}

	if(!create_ground(scn, sim)) {
		return false;
	}

	if(!create_galton(scn, sim)) {
		return false;
	}

	if(!create_balls(scn, sim)) {
		return false;
	}

	if(!create_exit_triggers(scn, sim)) {
		return false;
	}

	glEnable(GL_NORMALIZE);
	return true;
}

/* -- create_sky --
 * creates a low-poly sphere to which we're going to directly map the sky cubemap
 * using the sphere normals as texture coordinates (see sky_render func).
 */
static bool create_sky(scene *scn)
{
	texture *sky_cubemap = get_texture("sky.cube", TEX_CUBE);
	if(!sky_cubemap) {
		return false;
	}

	// showing the sky is a config option
	bool show_sky = true;
	cfg.getopt("sky", &show_sky);

	if(show_sky) {
		robject *sky = new robject;
		gen_sphere(sky->get_mesh(), 5000.0, 16, 8);
		sky->set_render_func(sky_render, 0);
		sky->get_material_ptr()->set_texture(sky_cubemap);
		scn->add_object(sky);
	}
	return true;
}

/* -- create_ground --
 * creates the ground plane renderable object and simulation quad
 */
static bool create_ground(scene *scn, Simulation *sim)
{
	double ground_y = -bp->row_dist * 2.0;

	// ground plane robject
	robject *obj = new robject;
	gen_plane(obj->get_mesh(), 5, 5);
	obj->set_scaling(Vector3(100.0, 100.0, 100.0));
	obj->set_position(Vector3(0, ground_y, 0));
	scn->add_object(obj);

	bool ground_tex = true;
	cfg.getopt("groundtex", &ground_tex);

	if(ground_tex) {
		material *mat = obj->get_material_ptr();
		mat->set_texture(get_texture("overlap.png"));
		mat->texture_xform()->set_scaling(Vector3(8, 16, 1));
	}

	// also create the ground plane simulation quad
	SimQuad *quad = new SimQuad;
	quad->v[0] = Vector3(-100.0, ground_y, 100.0);
	quad->v[1] = Vector3(100.0, ground_y, 100.0);
	quad->v[2] = Vector3(100.0, ground_y, -100.0);
	quad->v[3] = Vector3(-100.0, ground_y, -100.0);
	quad->normal = Vector3(0, 1, 0);
	quad->get_bounds();
	sim->add_static(quad);

	ground_sim_obj = quad;
	return true;
}

/* -- create_galton --
 * creates the galton box by calling create_galton_glass() and create_galton_pins_sides()
 * to create the glass panels and the pins & side quads respectively.
 */
static bool create_galton(scene *scn, Simulation *sim)
{
	if(!create_galton_glass(scn, sim)) {
		return false;
	}

	if(!create_galton_pins_sides(scn, sim)) {
		return false;
	}
	return true;
}

/* -- create_galton_glass --
 * creates the front/back glass panels as dual-sided slightly tesselated semi-transparent
 * quads. it also creates the corresponding simulation quads.
 */
static bool create_galton_glass(scene *scn, Simulation *sim)
{
	// glass pane material
	material mat;
	mat.set_color(color(0.5, 0.7, 1.0, 1));
	mat.set(0.2, MATTR_ALPHA);
	
	// glass pane geometry
	for(int i=0; i<2; i++) {
		robject *obj = new robject;
		gen_plane(obj->get_mesh(), 9, 3);
		trimesh mesh = *obj->get_mesh();
		mesh.flip_winding();
		mesh.flip_normals();
		obj->get_mesh()->merge(mesh);

		Vector3 pos = Vector3(0, bp->height / 2 - bp->pin_rad,
				(i ? bp->pin_length : -bp->pin_length) / 2.0);
		obj->set_position(pos);
		obj->set_scaling(Vector3(bp->width / 2, 1, bp->height / 2));
		obj->set_rotation(Vector3(-HALF_PI, 0, 0));
		scn->add_object(obj);

		obj->set_material(mat);

		// glass pane sim quad
		double x = bp->width / 2.0;

		// double layer :)
		for(int j=0; j<2; j++) {
			SimQuad *quad = new SimQuad;
			quad->v[0] = Vector3(-x, -ball_rad, pos.z);
			quad->v[1] = Vector3(x, -ball_rad, pos.z);
			quad->v[2] = Vector3(x, bp->height + 20, pos.z);
			quad->v[3] = Vector3(-x, bp->height + 20, pos.z);
			quad->normal = i ? Vector3(0, 0, -1) : Vector3(0, 0, 1);
			quad->get_bounds();
			sim->add_static(quad);

			pos.z += i ? 1.0 : -1.0;
		}
	}
	return true;
}

/* -- create_galton_pins_sides --
 * calculates pin positions, and creates the geometry and simulation cylinders for
 * each pin of the galton box (all sizes, counts, etc are configurable).
 * This function also creates the solid zig-zag quads along the sides, by connecting
 * the leftmost and rightmost pins.
 */
static bool create_galton_pins_sides(scene *scn, Simulation *sim)
{
	texture *dark_sky_cubemap = get_texture("dark_sky.cube", TEX_CUBE);
	if(!dark_sky_cubemap) {
		return false;
	}

	// setup pin material
	material pin_mat;
	pin_mat.set_color(color(0.2, 0.2, 0.2, 1.0));
	pin_mat.set_color(color(1.0, 1.0, 1.0, 1.0), MATTR_SPECULAR);
	pin_mat.set(80.0, MATTR_SHININESS);
	pin_mat.set_texture(dark_sky_cubemap);

	/* these arrays will hold the first and last pin positions of each row
	 * they are used to setup the zig-zag side quads later
	 */
	Vector3 *left_side = new Vector3[bp->rows + 1];
	Vector3 *right_side = new Vector3[bp->rows + 1];

	for(int i=0; i<bp->rows; i++) {
		int row_pins = (i & 1) ? bp->cols - 1 : bp->cols;

		for(int j=0; j<row_pins; j++) {
			double x = ((double)j - 0.5 * (double)(row_pins - 1)) * bp->pin_dist;

			Vector3 cylpos = Vector3(x, (double)((bp->rows - 1) - i) * bp->row_dist, 0);

			if(j == 0) {
				left_side[i + 1] = cylpos;
			} else if(j == row_pins - 1) {
				right_side[i + 1] = cylpos;
			}

			// pin renderable objects
			robject *obj = new robject;
			gen_cylinder(obj->get_mesh(), bp->pin_rad, 16, 4);
			obj->set_rotation(Vector3(HALF_PI, 0, 0));
			obj->set_scaling(Vector3(1, bp->pin_length, 1));
			obj->set_position(cylpos);
			obj->set_material(pin_mat);
			obj->set_render_func(envball_render, 0);
			scn->add_object(obj);

			// pin simulation cylinders
			SimCylinder *cyl = new SimCylinder;
			cyl->pos = Vector2(cylpos.x, cylpos.y);
			cyl->rad = bp->pin_rad;
			cyl->min_z = -bp->pin_length / 2.0;
			cyl->max_z = bp->pin_length / 2.0;
			cyl->get_bounds();
			sim->add_static(cyl);
		}
	}

	left_side[0] = left_side[1] + Vector3(0, 10, 0);
	right_side[0] = right_side[1] + Vector3(0, 10, 0);

	// side quads
	for(int i=1; i<bp->rows + 1; i++) {
		for(int j=0; j<2; j++) {
			Vector3 pin0 = j ? right_side[i - 1] : left_side[i - 1];
			Vector3 pin1 = j ? right_side[i] : left_side[i];
			float z = bp->pin_length / 2.0;

			Vector3 side_tang = (pin1 - pin0).normalized();
			float side_angle = acos(side_tang.x);

			// renderable quads
			robject *obj = new robject;
			gen_plane(obj->get_mesh(), 1, 1);
			trimesh mesh = *obj->get_mesh();
			mesh.flip_winding();
			mesh.flip_normals();
			obj->get_mesh()->merge(mesh);

			obj->set_position((pin0 + pin1) / 2.0);
			obj->set_rotation(Vector3(0, 0, side_angle));
			obj->set_scaling(Vector3((pin1 - pin0).length() / 2.0, 1.0, z));
			scn->add_object(obj);

			// collision quads
			SimQuad *quad = new SimQuad;
			quad->v[0] = pin0 + Vector3(0, 0, z);
			quad->v[1] = pin1 + Vector3(0, 0, z);
			quad->v[2] = pin1 + Vector3(0, 0, -z);
			quad->v[3] = pin0 + Vector3(0, 0, -z);
			quad->normal = (pin0 - pin1).rotated(Vector3(0, 0, j ? HALF_PI : -HALF_PI));
			quad->normal.normalize();
			quad->get_bounds();
			sim->add_static(quad);
		}
	}

	delete [] left_side;
	delete [] right_side;
	return true;
}

/* -- create_balls --
 * creates the renderable ball objects, and assigns appropriate materials for
 * the rubber, wooden and metal balls. the wooden ball is left untextured on purpose
 * since any wood grain on the balls would make it painfully obvious that the balls aren't
 * spinning at all as they go through the galton box :)
 */
static bool create_balls(scene *scn, Simulation *sim)
{
	texture *dark_sky_cubemap = get_texture("dark_sky.cube", TEX_CUBE);
	if(!dark_sky_cubemap) {
		return false;
	}

	// ball geometry
	for(int i=0; i<NUM_BALL_TYPES; i++) {
		robject *obj = new robject;
		gen_sphere(obj->get_mesh(), ball_rad, 20, 10);
		obj->set_position(Vector3(i * ball_rad * 2.0, ball_rad, 0));
		ball[i] = obj;
	}

	// setup ball materials
	material *mat_rubber = ball[BTYPE_RUBBER]->get_material_ptr();
	material *mat_wood = ball[BTYPE_WOOD]->get_material_ptr();
	material *mat_metal = ball[BTYPE_METAL]->get_material_ptr();

	mat_rubber->set_color(color(0.2, 0.2, 0.2, 1.0));
	mat_rubber->set_color(color(0.0, 0.0, 0.0, 1.0), MATTR_SPECULAR);

	mat_wood->set_color(color(0.59, 0.423, 0.211, 1.0));
	mat_wood->set_color(color(0.6, 0.6, 0.6, 1.0), MATTR_SPECULAR);
	mat_wood->set(60.0, MATTR_SHININESS);

	mat_metal->set_color(color(0.2, 0.2, 0.2, 1.0));
	mat_metal->set_color(color(1.0, 1.0, 1.0, 1.0), MATTR_SPECULAR);
	mat_metal->set(80.0, MATTR_SHININESS);
	mat_metal->set_texture(dark_sky_cubemap);

	// the metal balls need a custom render function for the cubemap reflection
	ball[BTYPE_METAL]->set_render_func(envball_render, 0);

	// add the balls render function in the scene
	scn->add_render_func(render_func(draw_balls, sim));
	return true;
}

static bool create_exit_triggers(scene *scn, Simulation *sim)
{
	int num_exits = bp->cols - 2;
	int *exit_cnt = new int[num_exits];
	memset(exit_cnt, 0, num_exits * sizeof *exit_cnt);

	ball_stats.num_exits = num_exits;
	ball_stats.exit_counters = exit_cnt;

	pthread_mutex_init(&ball_stats.mutex, 0);


	for(int i=0; i<num_exits; i++) {
		double y = -ball_rad;
		double x = bp->pin_dist / 2.0 - ball_rad;
		double z = bp->pin_length / 2.0;

		double xoffs = bp->pin_dist * (double)(i - num_exits / 2) + bp->pin_dist / 2.0;

		SimTrigger *quad = new SimTrigger;
		quad->v[0] = Vector3(-x + xoffs, y, z);
		quad->v[1] = Vector3(x + xoffs, y, z);
		quad->v[2] = Vector3(x + xoffs, y, -z);
		quad->v[3] = Vector3(-x + xoffs, y, -z);
		quad->normal = Vector3(0, 1, 0);
		quad->set_collision_callback(ball_exit_callback, exit_cnt + i);
		quad->get_bounds();
		sim->add_static(quad);

		// for each slot also add a render function object to draw the counter
		scn->add_render_func(render_func(draw_slot_counter, exit_cnt + i));
	}
	return true;
}

robject *get_ball_object(int btype)
{
	return ball[btype];
}

double get_ball_mass(int btype)
{
	return ball_mass[btype];
}

double get_ball_restitution(int btype)
{
	return ball_rest[btype];
}

double get_ball_radius(int btype)
{
	return ball_rad;
}

void drop_balls(Simulation *sim, int count, unsigned int type)
{
	float x_range = 0.85 * bp->width - ball_rad * 2.0;
	float z_range = bp->pin_length - ball_rad * 2.0;

	for(int i=0; i<count; i++) {
		SimBall *ball = new SimBall;

		std::list<Contact> tmp_list;
		do {
			float x = frand(x_range) - x_range / 2.0;
			float y = ball_rad + bp->height + frand(15.0);
			float z = frand(z_range) - z_range / 2.0;

			ball->pos = Vector3(x, y, z);
			tmp_list.clear();
		} while(sim->find_contacts(ball, &tmp_list) > 0);

		ball->rad = ball_rad;

		if(type == BTYPE_ANY) {
			ball->type = rand() % 3;
		} else {
			ball->type = type;
		}
		ball->mass = ball_mass[ball->type];
		ball->rest = ball_rest[ball->type];

		sim->add_ball(ball);

		pthread_mutex_lock(&ball_stats.mutex);
		ball_stats.num_spawned++;
		pthread_mutex_unlock(&ball_stats.mutex);

		send_spawn_message();
	}
}

static bool ball_exit_callback(const SimStaticObj *sobj, const SimBall *ball,
		const Contact *cont, void *cls)
{
	pthread_mutex_lock(&ball_stats.mutex);
	ball_stats.num_collected++;
	(*(int*)cls)++;	// increase the exit counter for this slot
	pthread_mutex_unlock(&ball_stats.mutex);

	// if we're a client, send an MSG_COLLECT
	send_collect_msg((int*)cls - ball_stats.exit_counters);

	return false;	// & don't respond to this collision
}
