#include <errno.h>
#include "pthread.h"
#include "impl.h"

#define VALID_THREAD(tid)	((tid) < MAX_THREADS && threads[tid].thread)

static void free_thread(pthread_t tid);
static DWORD WINAPI thread_func(void *arg);

#define MAX_THREADS		128
static struct thread_data threads[MAX_THREADS];
static int next_free;
static pthread_mutex_t tbl_mutex = PTHREAD_MUTEX_INITIALIZER;

struct thread_data *_pthread_get_data(pthread_t pid)
{
	if(pid < 0 || pid >= MAX_THREADS) {
		return 0;
	}
	return threads + pid;
}

int pthread_create(pthread_t *tid, const pthread_attr_t *attr,
		void *(*tfunc)(void*), void *arg)
{
	int prev_next_free = next_free;
	size_t ssize = attr ? attr->stack_size : 0;
	struct thread_data *tdata;

	/* allocate a slot in the thread table */
	pthread_mutex_lock(&tbl_mutex);
	if(next_free >= MAX_THREADS) {
		pthread_mutex_unlock(&tbl_mutex);
		errno = EAGAIN;
		return -1;
	}
	*tid = next_free;	/* return the pthread_t index */
	tdata = threads + next_free;

	while(++next_free < MAX_THREADS) {
		if(!threads[next_free].thread) {
			break;
		}
	}

	tdata->tfunc = tfunc;
	tdata->arg = arg;
	tdata->detached = attr ? attr->detached : 0;
	tdata->cancel_state = PTHREAD_CANCEL_ENABLE;
	tdata->cancel_type = PTHREAD_CANCEL_DEFERRED;

	pthread_mutex_init(&tdata->mut, 0);

	/* this event is used to receive cancellation notifications
	 * in pthread cancellation points.
	 */
	if(!(tdata->cancel_event = CreateEvent(0, 0, 0, 0))) {
		next_free = prev_next_free;
		pthread_mutex_unlock(&tbl_mutex);
		errno = EAGAIN;
		return -1;
	}

	/* create the thread */
	if(!(tdata->thread = CreateThread(0, ssize, thread_func, tdata, 0, &tdata->thread_id))) {
		next_free = prev_next_free;
		pthread_mutex_unlock(&tbl_mutex);
		CloseHandle(tdata->cancel_event);
		errno = EAGAIN;
		return -1;
	}
	pthread_mutex_unlock(&tbl_mutex);

	return 0;
}

static void free_thread(pthread_t tid)
{
	memset(threads + tid, 0, sizeof *threads);

	pthread_mutex_lock(&tbl_mutex);
	if(tid < next_free) {
		next_free = tid;
	}
	pthread_mutex_unlock(&tbl_mutex);
}

static DWORD WINAPI thread_func(void *arg)
{
	struct thread_data *t = arg;
	t->exit_status = t->tfunc(t->arg);

	_pthread_tsd_cleanup();

	if(t->detached) {
		int idx = t - threads;
		free_thread(idx);
	}
	return 0;
}

void pthread_exit(void *val_ptr)
{
	pthread_t tid;

	_pthread_tsd_cleanup();

	tid = pthread_self();
	if(threads[tid].detached) {
		free_thread(tid);
	} else {
		threads[tid].exit_status = val_ptr;
	}

	ExitThread(0);
}


/* TODO: detect EDEADLK */
int pthread_join(pthread_t tid, void **val_ptr)
{
	struct thread_data *t = threads + tid;

	if(!VALID_THREAD(tid)) {
		errno = ESRCH;
		return -1;
	}
	if(t->detached) {	/* not joinable */
		errno = EINVAL;
		return -1;
	}

	if(WaitForSingleObject(t->thread, INFINITE) != WAIT_OBJECT_0) {
		errno = EINVAL;
		return -1;
	}

	if(val_ptr) {
		*val_ptr = t->exit_status;
	}

	free_thread(tid);
	return 0;
}

int pthread_detach(pthread_t tid)
{
	if(!VALID_THREAD(tid)) {
		errno = ESRCH;
		return -1;
	}
	if(threads[tid].detached) {
		errno = EINVAL;
		return -1;
	}
	threads[tid].detached = 1;
	return 0;
}

pthread_t pthread_self(void)
{
	int i;
	DWORD thread_id = GetCurrentThreadId();
		
	for(i=0; i<MAX_THREADS; i++) {
		if(threads[i].thread_id == thread_id) {
			return i;
		}
	}

	return -1;
}

int pthread_equal(pthread_t t1, pthread_t t2)
{
	return t1 == t2;
}


int pthread_attr_init(pthread_attr_t *attr)
{
	memset(attr, 0, sizeof *attr);
	return 0;
}

int pthread_attr_destroy(pthread_attr_t *attr)
{
	return 0;
}

int pthread_attr_getdetachstate(const pthread_attr_t *attr, int *dstate)
{
	*dstate = attr->detached;
	return 0;
}

int pthread_attr_setdetachstate(pthread_attr_t *attr, int dstate)
{
	attr->detached = dstate;
	return 0;
}

int pthread_attr_getguardsize(const pthread_attr_t *attr, size_t *gsize)
{
	*gsize = attr->guard_size;
	return 0;
}

int pthread_attr_setguardsize(pthread_attr_t *attr, size_t gsize)
{
	attr->guard_size = gsize;
	return 0;
}

int pthread_attr_getschedparam(const pthread_attr_t *attr, struct sched_param *param)
{
	return 0;	/* TODO */
}

int pthread_attr_setschedparam(pthread_attr_t *attr, const struct sched_param *param)
{
	return 0;	/* TODO */
}

int pthread_attr_getschedpolicy(const pthread_attr_t *attr, int *policy)
{
	*policy = attr->sched_policy;
	return 0;
}

int pthread_attr_setschedpolicy(pthread_attr_t *attr, int policy)
{
	attr->sched_policy = policy;
	return 0;
}

int pthread_attr_getinheritsched(const pthread_attr_t *attr, int *inherit)
{
	return 0;	/* TODO */
}

int pthread_attr_setinheritsched(pthread_attr_t *attr, int inherit)
{
	return 0;	/* TODO */
}

int pthread_attr_getscope(const pthread_attr_t *attr, int *scope)
{
	*scope = attr->scope;
	return 0;
}

int pthread_attr_setscope(pthread_attr_t *attr, int scope)
{
	attr->scope = scope;
	return 0;
}

int pthread_attr_getstacksize(const pthread_attr_t *attr, size_t *ssize)
{
	*ssize = attr->stack_size;
	return 0;
}

int pthread_attr_setstacksize(pthread_attr_t *attr, size_t ssize)
{
	attr->stack_size = ssize;
	return 0;
}

int pthread_setschedprio(pthread_t tid, int prio)
{
	return 0;	/* TODO */
}
