#ifndef SIMOBJ_H_
#define SIMOBJ_H_

#include <vmath.h>
#include <henge.h>

enum {
	BTYPE_RUBBER,
	BTYPE_WOOD,
	BTYPE_METAL,

	NUM_BALL_TYPES
};
#define BTYPE_ANY		NUM_BALL_TYPES
#define BTYPE_INVALID	0xffff


class SimBall {
public:
	bool sim;
	double rad;
	Vector3 pos, vel, forces;
	double mass;
	double rest;

	unsigned int type;

	Vector3 prev_vel;
	Vector3 end;	// calculated as pos + vel * dt, used in collision det.

	SimBall(const Vector3 &pos = Vector3(0,0,0), double rad = 1.0, double mass = 1.0, double rest = 1.0);
};

class Contact;

bool ball_intersect(const SimBall *ball1, const SimBall *ball2, Contact *contact);
bool ball_contact(const SimBall *ball1, const SimBall *ball2, Contact *contact);

struct AABox {
	Vector3 min, max;
	bool valid;
};

class SimStaticObj {
protected:
	void *col_func_cls;
	bool (*collision_func)(const SimStaticObj*, const SimBall*, const Contact *, void*);

	bool active;
	mutable AABox aabox;

	virtual void calc_bounds() = 0;

public:
	SimStaticObj();
	virtual ~SimStaticObj();

	virtual void activate();
	virtual void deactivate();
	inline bool is_active() const { return active; }

	virtual const AABox &get_bounds() const;

	virtual void set_collision_callback(bool (*func)(const SimStaticObj*,
				const SimBall*, const Contact*, void*), void *cls = 0);
	virtual bool signal_collision(const SimBall *ball, const Contact *cont) const;

	virtual bool intersect(const SimBall *ball, Contact *contact) const = 0;
	virtual bool contact(const SimBall *ball, Contact *contact) const = 0;
	virtual void draw(const henge::color &col) const = 0;
};

class SimQuad : public SimStaticObj {
protected:
	virtual void calc_bounds();

public:
	Vector3 v[4];
	Vector3 normal;

	SimQuad();
	SimQuad(const Vector3 &a, const Vector3 &b, const Vector3 &c, const Vector3 &d);
	virtual ~SimQuad();

	virtual bool intersect(const SimBall *ball, Contact *contact) const;
	virtual bool contact(const SimBall *ball, Contact *contact) const;
	virtual void draw(const henge::color &col) const;
};

class SimTrigger : public SimQuad {
public:
	virtual bool contact(const SimBall *ball, Contact *contact) const;
};

class SimCylinder : public SimStaticObj {
protected:
	virtual void calc_bounds();

public:
	Vector2 pos;
	double rad, min_z, max_z;

	SimCylinder();
	SimCylinder(Vector2 pos, double rad);
	virtual ~SimCylinder();

	virtual bool intersect(const SimBall *ball, Contact *contact) const;
	virtual bool contact(const SimBall *ball, Contact *contact) const;
	virtual void draw(const henge::color &col) const;
};

class Contact {
public:
	double t;
	Vector3 pos, normal;
	SimBall *ball1, *ball2;
	SimStaticObj *sobj;

	// used for sorting contacts
	inline bool operator <(const Contact &c2) const { return t < c2.t; }
};

#endif	// SIMOBJ_H_
